'use strict';

var fs = require('fs');
var net = require('net');

var app = require('../modules/app.js');
var ExpectedError = require('../modules/errors.js').ExpectedError;

var EXTRACT_APP_SCRIPT_PATH = '/tmp/extract_app.sh';
var DEFAULT_KILLING_TIMEOUT = 5000;

var appLogFileHandle;
var spawn = require('child_process').spawn;

try {
    appLogFileHandle = uv.fs_open(app.logPath, 'rw', parseInt('644', 8), 0);
} catch (error) {
    appLogFileHandle = -1;

    console.error('Error opening application log file.');
    console.error(error);
}

exports.start = function (options, callback) {
    switch (app.status) {
        case 'starting':
        case 'started':
            callback(new ExpectedError('Application has already been started', 3));
            return;
        case 'not-deployed':
            callback(new ExpectedError('Application has not been deployed', 1));
            return;
    }

    var args = [
        app.launcherPath,
        '--console'
    ];

    if (options.debug) {
        args.push('--debugger');
    }

    try {
        console.log('Starting application launcher with arguments `' + args.join(' ') + '` in "' + app.path + '".');

        var handle = spawn(uv.exepath(), args, {
            cwd: app.path,
            stdio: ['ignore', appLogFileHandle, appLogFileHandle]
        });

        handle.on('exit', function (code) {
            app.updateStatus('not-started');
            app.removeListener('started', onstarted);
            console.log('Application exited with code ' + code + '.');
        });

        app.updateStatus('starting', handle);
        app.once('started', onstarted);
    } catch (error) {
        console.error('Application failed to start.');
        callback(new ExpectedError('Application failed to start', 2, error));
    }

    function onstarted() {
        callback();
    }
};

exports.stop = function (options, callback) {
    switch (app.status) {
        case 'not-deployed':
            callback(new ExpectedError('No application has been deployed yet', 1));
            return;
        case 'not-started':
            callback(new ExpectedError('Application has not been started yet', 3));
            return;
    }

    var stopEventListenersCount = app.listenerCount('stop');

    console.log('stopEventListenersCount', stopEventListenersCount);

    app.once('stop', onstop);

    if (stopEventListenersCount > 0) {
        startTimer();
        return;
    }

    var timer;

    try {
        console.log('Stopping application...');

        app.handle.kill('SIGINT');

        startTimer();
    } catch (error) {
        console.error('Failed to stop application.');
        callback(new ExpectedError('Failed to stop application', 2, error));
    }

    function onstop() {
        clearTimeout(timer);
        console.log('Application exited.');

        app.updateStatus('not-started');

        callback();
    }

    function startTimer() {
        timer = setTimeout(function () {
            // TODO: redeunt app status checking?

            console.log('Killing application...');

            try {
                app.handle.kill('SIGKILL');
            } catch (e) { }
        }, DEFAULT_KILLING_TIMEOUT);
    }
};

exports.deploy = function (options, callback) {
    switch (app.status) {
        case 'starting':
        case 'started':
            callback(new ExpectedError('Application is currently running', 1));
            return;
    }

    download();

    function download() {
        var remote = options.remote;
        var stream = options.stream;

        if (stream) {
            console.log('Start downloading application package...');
        } else {
            console.log('Start downloading application package from ' + remote.ip + ':' + options.port + '...');

            stream = net.connect(options.port, remote.ip, function () {
                console.log('Connected to rap server.');
            });
        }

        var downloaded = 0;
        var hasError = false;

        fs.writeFileSync(app.deploymentPackagePath, '');

        stream.on('data', function (data) {
            downloaded += data.length;

            console.log('Downloaded ' + Math.floor(downloaded / 1024) + ' kB.');

            fs.appendFileSync(app.deploymentPackagePath, data);
            Duktape.gc();
        });

        stream.on('end', function () {
            console.log('Application package downloading socket ended.');

            if (hasError) {
                return;
            }

            extract();
        });

        stream.on('error', function (error) {
            hasError = true;
            callback(new ExpectedError('Failed to save application package', 2, error));
        });
    }

    function extract() {
        var script = '\
#!/bin/sh\n\
rm -fr ' + app.path + '\n\
mkdir -p ' + app.path + '\n\
cd ' + app.path + '\n\
tar -xf ' + app.deploymentPackagePath + '\n\
sync\n\
';

        fs.writeFile(EXTRACT_APP_SCRIPT_PATH, script, {
            mode: parseInt('755', 8)
        }, function (error) {
            if (error) {
                callback(new ExpectedError('Failed to extract application package', 2));
                return;
            }

            try {
                var handle = spawn(EXTRACT_APP_SCRIPT_PATH, [], {
                    cwd: '/tmp',
                    stdio: ['ignore', 'ignore', 'ignore']
                });

                handle.on('exit', function (code) {
                    if (code === 0) {
                        app.updateStatus('not-started');
                        callback();
                    } else {
                        callback(new ExpectedError('Failed to extract application package', 2));
                    }
                });
            } catch (error) {
                callback(new ExpectedError('Failed to extract application package', 2, error));
            }
        });
    }
};
