'use strict';

var net = require('net');
var path = require('path');
var util = require('util');

var home = require('home');

var errors = require('./errors.js');
var md5 = require('../libs/md5.js');
var packet = require('./packet.js');
var storage = require('./storage.js').default;

var Server = home.Server;
var HomeResponse = home.Response;

var ExpectedError = errors.ExpectedError;

var UNKNOWN_ERROR_STATUS = errors.UNKNOWN_ERROR_STATUS;
var UNKNOWN_COMMAND_STATUS = errors.UNKNOWN_COMMAND_STATUS;
var WRONG_SIGNATURE_STATUS = errors.WRONG_SIGNATURE_STATUS;
var SIGNATURE_REQUIRED_STATUS = errors.SIGNATURE_REQUIRED_STATUS;

var hop = Object.prototype.hasOwnProperty;

var serverConfig = require('../config.json').server;

var VERSION = require('../../package.json').version;
var PING_INTERVAL = 5000;
var EXECUTION_TIMEOUT = 10000;

// TODO: remove this mapping in the future and update rap with new commands.

/* eslint-disable quote-props */

var commandTypeMapping = {
    'deploy': 'app.deploy',
    'start': 'app.start',
    'stop': 'app.stop',
    'ping': 'system.ping',
    'upgrade': 'system.upgrade',
    'rename': 'system.rename',
    'reset': 'system.reset',
    'reboot': 'system.reboot',
    'set-wifi': 'system.setWiFi',
    'system.setWifi': 'system.setWiFi',
    'set-ap': 'system.setAP',
    'get-net-status': 'system.getNetStatus',
    'get-system-info': 'system.getInfo'
};

var bypassAuthCommandSet = {
    'system.preflight': true,
    'system.ping': true
};

/* eslint-enable quote-props */

var commandModuleMap = {
    app: require('../commands/app.js'),
    system: require('../commands/system.js')
};

var pendingCommandInfos = [];
var executing = false;
var executionTimer;

exports.setupSocketServer = function (logReader) {
    var clients = [];

    var server = net.createServer(function (client) {
        var address;

        try {
            address = client._getpeername();
        } catch (error) {
            console.error(error);
            client.destroy();
            return;
        }

        console.log('Client ' + address.ip + ' connected.');

        var parser = new packet.Parser();

        parser.on('data', function (command, stream) {
            appendCommand(command, {
                auth: 'signature',
                stream: stream,
                address: address
            }, client);
        });

        parser.on('error', function (error) {
            console.error('Parser error.');
            console.error(error);
        });

        client.on('data', function (data) {
            // TODO: there was a bug in ruff_module's net, now it's fixed.
            // This should be removed from ruffd and rap in the future.
            if (data.length === 5 && data.toString('binary') === 'close') {
                removeClient();
                client.destroy();
            } else {
                parser.append(data);
            }
        });

        client.on('close', function () {
            console.log('Client ' + address.ip + ' disconnected.');
        });

        client.on('error', function (error) {
            removeClient();
            console.error('Client ' + address.ip + ' connection error.');
            console.trace(error);
        });

        clients.push(client);

        function removeClient() {
            var index = clients.indexOf(client);

            if (index >= 0) {
                clients.splice(index, 1);
            }

            console.log('Client ' + address.ip + ' removed.');
        }
    });

    server.on('error', function (error) {
        console.error('Socket server error occurs.');
        console.error(error);
    });

    server.listen(serverConfig.port, function () {
        console.log('Ruffd socket server is now listening on port ' + serverConfig.port + '...');
    });

    setInterval(function () {
        var pingPacket = packet.buildPingPacket();

        clients.forEach(function (client) {
            try {
                client.write(pingPacket);
            } catch (error) {
                console.error('Failed to ping client.');
                console.error(error);
            }
        });
    }, PING_INTERVAL);

    logReader.on('log', function (text) {
        var logPacket = packet.buildLogPacket('normal', text);

        clients.forEach(function (client) {
            try {
                client.write(logPacket);
            } catch (error) {
                console.error('Failed to send log.');
                console.error(error);
            }
        });
    });

    logReader.start();
};

exports.setupHttpServer = function () {
    var server = new Server({
        views: path.join(__dirname, '../../views')
    });

    server.use('/', Server.static(path.join(__dirname, '../../static')));

    server.use('/', require('./middleware/json-parser'));
    server.use('/', require('./middleware/x-body-parser'));

    server.use('/', function (req, res) {
        res.setHeader('Access-Control-Allow-Origin', req.headers['origin'] || '*');
        res.setHeader('Access-Control-Allow-Methods', req.headers['access-control-request-method'] || '*');
        res.setHeader('Access-Control-Allow-Headers', req.headers['access-control-request-headers'] || '*');
        return req;
    });

    server.use('/', function (req) {
        if (req.method === 'OPTIONS') {
            return undefined;
        } else {
            return req;
        }
    });

    var port = storage.get('http-port') || 80;

    server.get('/', function (req) {
        return {
            version: VERSION
        };
    });

    server.post('/api/call', function (req) {
        return new Promise(function (resolve, reject) {
            var headers = req.headers;

            appendCommand(req.body, {
                auth: 'basic',
                authorization: headers['authorization'],
                stream: headers['content-type'] === 'application/stream-payload' ?
                    req : undefined,
                address: req.socket.remoteAddress
            }, function (error, result) {
                if (error) {
                    reject(error);
                    return;
                }

                resolve({
                    result: result
                });
            });
        })
            .catch(function (reason) {
                if (reason instanceof AuthorizationRequiredError) {
                    return new AuthorizationRequiredResponse();
                }

                var data = {
                    error: {
                        message: reason.message
                    }
                };

                if (reason instanceof ExpectedError) {
                    console.log(reason.message);
                } else if (reason) {
                    console.error(reason.stack);
                }

                return data;
            });
    });

    server
        .listen(port)
        .then(function () {
            console.log('Ruffd HTTP server is now listening on port ' + port + '...');
        });
};

function AuthorizationRequiredError() {
    ExpectedError.apply(this, arguments);
}

util.inherits(AuthorizationRequiredError, ExpectedError);

function AuthorizationRequiredResponse() { }

util.inherits(AuthorizationRequiredResponse, HomeResponse);

AuthorizationRequiredResponse.prototype.applyTo = function (res) {
    res.statusCode = 401;
    res.end();
};

var BASIC_AUTH_PREFIX = 'Basic ';

function verifyAuthorization(authorization, type) {
    var token = storage.get('auth-token');

    if (
        !token ||
        token === 'd41d8cd98f00b204e9800998ecf8427e' /* md5('') */ ||
        hop.call(bypassAuthCommandSet, type)
    ) {
        return;
    }

    if (authorization && authorization.indexOf(BASIC_AUTH_PREFIX) === 0) {
        authorization = new Buffer(authorization.substr(BASIC_AUTH_PREFIX.length), 'base64').toString();

        var parts = authorization.split(':');
        var password = parts[1] || parts[0];

        var calculatedToken = md5(password);

        if (calculatedToken === token) {
            return;
        }
    }

    throw new AuthorizationRequiredError();
}

function verifySignature(command, type) {
    var token = storage.get('auth-token');

    if (!token || hop.call(bypassAuthCommandSet, type)) {
        return;
    }

    var signature = command.signature;

    if (!signature) {
        throw new ExpectedError('Signature expected', SIGNATURE_REQUIRED_STATUS);
    }

    delete command.signature;

    command.token = token;

    var dataArray = Object
        .keys(command)
        .filter(function (key) {
            // This should be redundant, but let's make sure.
            return command[key] !== undefined;
        })
        .sort()
        .map(function (key) {
            return [key, command[key]];
        });

    var calculatedSignature = md5(JSON.stringify(dataArray));

    if (calculatedSignature !== signature) {
        throw new ExpectedError('Wrong signature', WRONG_SIGNATURE_STATUS);
    }
}

/**
 * @param {Object} command
 * @param {Object} options
 * @param {Socket | Function} client client object or a callback
 */
function appendCommand(command, options, client) {
    var type = command.type;

    var callback;

    if (typeof client === 'function') {
        callback = client;
        client = undefined;
    } else {
        callback = createCallback(command, client);
    }

    if (hop.call(commandTypeMapping, type)) {
        type = commandTypeMapping[type];
    }

    try {
        if (options.auth === 'basic') {
            verifyAuthorization(options.authorization, type);
        } else {
            verifySignature(command, type);
        }
    } catch (error) {
        callback(error);
        return;
    }

    var parts = type.split('.');

    var category = parts[0];
    var operationName = parts[1];

    var commandModule = hop.call(commandModuleMap, category) ?
        commandModuleMap[category] : undefined;

    var operation = commandModule && hop.call(commandModule, operationName) ?
        commandModule[operationName] : undefined;

    command.remote = options.address;
    command.stream = options.stream;

    if (operation) {
        pendingCommandInfos.push({
            operation: operation.bind(commandModule),
            command: command,
            callback: callback
        });

        nextCommand();
    } else {
        callback(new ExpectedError('Unknown command "' + type + '"', UNKNOWN_COMMAND_STATUS));
    }
}

function nextCommand() {
    if (executing) {
        return;
    }

    var info = pendingCommandInfos.shift();

    if (!info) {
        return;
    }

    executing = true;

    var callback = info.callback;

    info.operation(info.command, function (error, result) {
        clearTimeout(executionTimer);
        preNextCommand();
        callback(error, result);
    });

    // This is not going to timeout current executing command, but just start
    // the next command so it won't wait too long.
    executionTimer = setTimeout(preNextCommand, EXECUTION_TIMEOUT);

    function preNextCommand() {
        executing = false;
        nextCommand();
    }
}

function createCallback(command, client) {
    var called = false;

    return function (error, value) {
        if (called) {
            console.error('Command "' + command.type + '" callback has been called more than once.');
            return;
        }

        console.log('Command "' + command.type + '" callback called.');

        if (error instanceof ExpectedError) {
            console.log('[ExpectedError]', error.message);
        } else if (error) {
            console.error(error.stack);
        }

        called = true;

        var ackPacket = packet.buildAckPacket(
            command.id,
            value,
            error ? error.code || UNKNOWN_ERROR_STATUS : 0
        );

        try {
            client.write(ackPacket);
        } catch (error) {
            console.error('Failed to send back ack.');
            console.error(error);
        }
    };
}
