'use strict';

var EventEmitter = require('events');
var util = require('util');

/* eslint-disable quote-props */
var ESCAPE_ENCODE_LIST = {
    '03': '%03',  // ctrl-c
    '25': '%25',  // %
    '0d': '%0d',  // 0x0d
    '0a': '%0a',  // 0x0a
    '1a': '%1a',  // 0x1a ctrl-z
    '11': '%11',  // 0x11
    '12': '%12',  // 0x12
    '13': '%13',  // 0x13
    '14': '%14',  // 0x14
    '1c': '%1c'   // 0x1c
};
/* eslint-enable quote-props */

var ESCAPE_DECODE_LIST = {
    '%03': 0x03,
    '%25': 0x25,
    '%0d': 0x0d,
    '%0a': 0x0a,
    '%1a': 0x1a,
    '%11': 0x11,
    '%12': 0x12,
    '%13': 0x13,
    '%14': 0x14,
    '%1c': 0x1c
};

var ESCAPE_PREFIX = new Buffer('%');
var TIMEOUT = 10000;

function Escape() {
    EventEmitter.call(this);
    this._buffer = new Buffer(0);
    this._pendingLength = 0;
    this._pendingBuffer = new Buffer(0);
    this._timeout = TIMEOUT;
}
util.inherits(Escape, EventEmitter);

Escape.prototype.encode = function (data) {
    var i;
    var res = new Buffer(0);
    var str;
    var escaped;
    for (i = 0; i < data.length; i++) {
        str = data.slice(i, i + 1).toString('hex');
        escaped = ESCAPE_ENCODE_LIST[str];
        if (escaped !== undefined) {
            res = Buffer.concat([res, new Buffer(escaped)]);
        } else {
            res = Buffer.concat([res, data.slice(i, i + 1)]);
        }
    }
    return res;
};

Escape.prototype.pushUndecodedData = function (data) {
    clearTimeout(this._timeHandler);
    this._decode(data);
};

Escape.prototype._setTimer = function () {
    var that = this;
    this._timeHandler = setTimeout(function () {
        that._pendingBuffer = new Buffer(0);
        that.emit('error', new Error('data is incomplete'));
    }, this._timeout);
};

Escape.prototype._decode = function (data) {
    var i;
    var length = data.length;
    var res = new Buffer(0);
    var current;

    for (i = 0; i < length; i++) {
        current = data.slice(i, i + 1);
        if (this._pendingLength === 0) {
            if (current.equals(ESCAPE_PREFIX)) {
                this._pendingLength = 2;
                this._pendingBuffer = new Buffer(current);
            } else {
                res = Buffer.concat([res, current]);
            }
        } else if (this._pendingLength === 1) {
            this._pendingLength = 0;
            this._pendingBuffer = Buffer.concat([
                this._pendingBuffer,
                current
            ]);
            var decodedValue = ESCAPE_DECODE_LIST[this._pendingBuffer.toString('utf8')];
            if (decodedValue) {
                res = Buffer.concat([res, new Buffer([decodedValue])]);
            } else {
                res = Buffer.concat([res, new Buffer([0])]);
            }
        } else if (this._pendingLength === 2) {
            this._pendingLength = 1;
            this._pendingBuffer = Buffer.concat([
                this._pendingBuffer,
                current
            ]);
        }
    }
    if (this._pendingLength > 0) {
        this._setTimer();
    }

    this.emit('data', res);
};

module.exports = Escape;
